/*
 * JBoss, Home of Professional Open Source.
 * Copyright 2009, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags. See the copyright.txt file in the
 * distribution for a full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.infinispan.loaders.jpa;

import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.Id;
import javax.persistence.Persistence;

import org.infinispan.Cache;
import org.infinispan.container.entries.InternalCacheEntry;
import org.infinispan.container.entries.InternalEntryFactory;
import org.infinispan.loaders.CacheLoaderConfig;
import org.infinispan.loaders.CacheLoaderException;
import org.infinispan.loaders.LockSupportCacheStore;
import org.infinispan.marshall.Marshaller;

/**
 * {@link org.infinispan.loaders.CacheStore} implementation that stores the
 * entries in a JPA {@link javax.persistence.PersistenceContext}.
 * {@link javax.persistence.Entity} are loaded to cache and persisted when cache
 * requests. Cache key is generated using reflection based on
 * {@link javax.persistence.Id}.
 * <p/>
 * All files required (including JPA implementation) needs to be on classpath in
 * order to be able to instantiate the
 * {@link javax.persistence.EntityManagerFactory}.
 * <p/>
 * 
 * @author Rafael Ribeiro <rafaelri@gmail.com>
 * @see JPACacheStoreConfig
 */
public class JPACacheStore extends LockSupportCacheStore {
	private EntityManagerFactory emf;
	private JPACacheStoreConfig config;
	private Class entityClass;
	private Method idReadMethod;
	private Field idField;
	private String entityClassName;

	public void init(CacheLoaderConfig config, Cache cache, Marshaller m)
			throws CacheLoaderException {
		super.init(config, cache, m);
		this.config = ((JPACacheStoreConfig) config);
		this.entityClassName = this.config.getEntityClassName();
		try {
			this.entityClass = Class.forName(this.config.getEntityClassName());
			setupId();
		} catch (Exception e) {
			throw new CacheLoaderException(
					"Unable to initialize JPACacheStore", e);
		}
	}

	public void start() throws CacheLoaderException {
		super.start();
		// implement possibility of looking up thru jndi?
		emf = Persistence.createEntityManagerFactory(config
				.getPersistenceUnitName());
	}

	public void stop() throws CacheLoaderException {
		super.stop();
		emf.close();
	}

	protected void clearLockSafe() throws CacheLoaderException {
		EntityManager em = null;
		try {
			em = emf.createEntityManager();
			em.createQuery("delete from " + config.getEntityClassName())
					.executeUpdate();
		} finally {
			closeEntityManager(em);
		}
	}

	private void closeEntityManager(EntityManager em) {
		if (em != null)
			em.close();
	}

	protected void fromStreamLockSafe(ObjectInput ois)
			throws CacheLoaderException {
		try {
			int numEntries = (Integer) marshaller.objectFromObjectStream(ois);
			clearLockSafe();
			for (int i = 0; i < numEntries; i++) {
				InternalCacheEntry e = (InternalCacheEntry) marshaller
						.objectFromObjectStream(ois);
				store(e);
			}
		} catch (Exception e) {
			throw new CacheLoaderException(e);
		}

	}

	protected String getLockFromKey(Object key) throws CacheLoaderException {
		return entityClassName + ":" + key;
	}

	protected Set<InternalCacheEntry> loadAllLockSafe()
			throws CacheLoaderException {
		EntityManager em = null;
		try {
			em = emf.createEntityManager();
			Set<InternalCacheEntry> result = new HashSet<InternalCacheEntry>();
			// oracle does not support in clauses with more than 1000 elements
			int batchSize = config.getBatchSize();
			if (batchSize > 1000)
				batchSize = 1000;
			List keys = em.createQuery(
					"select e." + idField.getName() + " from "
							+ config.getEntityClassName() + " e").setHint(
					"org.hibernate.fetchSize", batchSize).setHint("fetchSize",
					batchSize).getResultList();
			while (keys.size() > 0) {
				List loadedKeys = getListHead(keys, batchSize);
				List loadedValues = em.createQuery(
						"select e from " + config.getEntityClassName()
								+ " e where e." + idField.getName()
								+ " in (:id)").setParameter("id", loadedKeys)
						.setHint("org.hibernate.fetchSize", batchSize).setHint(
								"fetchSize", batchSize).getResultList();
				for (Object o : loadedValues) {
					Object key = getKey(o);
					result.add(InternalEntryFactory.create(key, o));
				}
				em.clear();
			}
			return result;
		} finally {
			closeEntityManager(em);
		}
	}

	public List getListHead(List l, int count) {
		// create a new list in order to avoid concurrent modification exception
		List result = new ArrayList();
		result.addAll(l.subList(0, count > l.size() ? l.size() : count));
		l.removeAll(result);
		return result;
	}

	protected InternalCacheEntry loadLockSafe(Object key, String lock)
			throws CacheLoaderException {
		EntityManager em = null;
		try {
			em = emf.createEntityManager();
			return InternalEntryFactory.create(key, em.find(entityClass, key));
		} finally {
			closeEntityManager(em);
		}
	}

	protected boolean removeLockSafe(Object key, String lock)
			throws CacheLoaderException {
		EntityManager em = null;
		try {
			em = emf.createEntityManager();
			em.remove(em.getReference(entityClass, key));
			return true;
		} finally {
			closeEntityManager(em);
		}
	}

	protected void storeLockSafe(InternalCacheEntry entry, String lock)
			throws CacheLoaderException {
		EntityManager em = null;
		try {
			em = emf.createEntityManager();
			em.merge(entry.getValue());
		} finally {
			closeEntityManager(em);
		}
	}

	protected void toStreamLockSafe(ObjectOutput oos)
			throws CacheLoaderException {
		try {
			Set<InternalCacheEntry> all = loadAllLockSafe();
			marshaller.objectToObjectStream(all.size(), oos);
			for (InternalCacheEntry se : all)
				marshaller.objectToObjectStream(se, oos);
		} catch (Exception e) {
			throw new CacheLoaderException(e);
		}
	}

	protected void purgeInternal() throws CacheLoaderException {
		System.out.println("purgeInternal");

	}

	public Class<? extends CacheLoaderConfig> getConfigurationClass() {
		return JPACacheStoreConfig.class;
	}

	protected void setupId() throws IntrospectionException {
		for (Field field : entityClass.getDeclaredFields()) {
			if (field.getAnnotation(Id.class) != null) {
				idField = field;
				if (field.isAccessible()) {
					break;
				} else {
					idReadMethod = getReadMethod(entityClass, field.getName());
				}
			}
		}
		if (idField == null) {
			for (PropertyDescriptor pd : Introspector.getBeanInfo(entityClass)
					.getPropertyDescriptors()) {
				if (pd.getReadMethod().getAnnotation(Id.class) != null) {
					idReadMethod = pd.getReadMethod();
					for (Field field : entityClass.getDeclaredFields()) {
						if (field.getName().equals(
								Introspector.decapitalize(pd.getName()))) {
							idField = field;
						}
					}
					break;
				}
			}
		}
	}

	protected Method getReadMethod(Class clazz, String field)
			throws IntrospectionException {
		for (PropertyDescriptor pd : Introspector.getBeanInfo(clazz)
				.getPropertyDescriptors()) {
			if (field.equals(Introspector.decapitalize(pd.getName()))) {
				return pd.getReadMethod();
			}
		}
		return null;
	}

	public Object getKey(Object value) {
		try {
			if ((idField != null) && (idField.isAccessible())) {
				return idField.get(value);
			} else if (idReadMethod != null) {
				return idReadMethod.invoke(value, null);
			} else {
				throw new IllegalStateException(value.getClass()
						+ " id isn't accessible");
			}
		} catch (Exception e) {
			throw new RuntimeException("unable to lookup id on "
					+ value.getClass(), e);
		}
	}

}
